use std::{collections::HashMap, rc::Rc};

use crate::node::NodeEnum;

pub type EnvironmentInternalSymbolKey = u16;
// pub type Environment = HashMap<EnvironmentInternalSymbolKey, Rc<NodeEnum>>;

pub struct Environment {
    map: HashMap<EnvironmentInternalSymbolKey, Rc<NodeEnum>>,
    stack_depth: usize,
    max_stack_depth: usize,
    stack: Vec<Vec<EnvironmentInternalSymbolKey>>,

    symbol_to_id: HashMap<String, EnvironmentInternalSymbolKey>,
    id_to_symbol: HashMap<EnvironmentInternalSymbolKey, String>,
    unique_keys: EnvironmentInternalSymbolKey,
}

impl Environment {
    pub fn new() -> Self {
        Self {
            map: HashMap::new(),
            stack_depth: 0,
            max_stack_depth: 5000,
            stack: vec![vec![]],

            symbol_to_id: HashMap::new(),
            id_to_symbol: HashMap::new(),
            unique_keys: 0,
        }
    }

    pub fn push_stack(&mut self) -> Result<(), String> {
        debug_assert!(self.stack_depth == self.stack.len() - 1);

        if self.stack_depth == self.max_stack_depth {
            self.unwind_stack();
            return Err("Max Stack depth exceeded".to_owned());
        }

        self.stack_depth += 1;
        self.stack.push(vec![]);

        debug_assert!(self.stack_depth == self.stack.len() - 1);

        Ok(())
    }

    pub fn pop_stack(&mut self) -> Result<(), String> {
        debug_assert!(
            self.stack_depth == self.stack.len() - 1,
            "Expected stack len: {}, stack_depth: {}",
            self.stack.len(),
            self.stack_depth
        );

        if self.stack_depth == 0 {
            return Err("Trying to pop empty stack".to_owned());
        }

        self.stack_depth -= 1;

        for item in self.stack.pop().unwrap() {
            self.map.remove(&item);
        }

        debug_assert!(self.stack_depth == self.stack.len() - 1);

        Ok(())
    }

    fn unwind_stack(&mut self) {
        for stack in &self.stack {
            for item in stack {
                self.map.remove(&item);
            }
        }

        self.stack = vec![vec![]];
    }

    pub fn get(&self, key: &EnvironmentInternalSymbolKey) -> Option<&Rc<NodeEnum>> {
        self.map.get(key)
    }

    pub fn insert(&mut self, key: EnvironmentInternalSymbolKey, value: Rc<NodeEnum>) {
        self.map.insert(key, value);
        unsafe { self.stack.get_unchecked_mut(self.stack_depth).push(key) }
    }

    pub fn str_to_id(&self, value: &String) -> Option<&EnvironmentInternalSymbolKey> {
        self.symbol_to_id.get(value)
    }
    pub fn id_to_str(&self, value: &EnvironmentInternalSymbolKey) -> Option<&String> {
        self.id_to_symbol.get(value)
    }
    pub fn insert_str_to_id(&mut self, key: String, value: EnvironmentInternalSymbolKey) {
        self.symbol_to_id.insert(key.clone(), value);
        self.id_to_symbol.insert(value, key);
    }
    pub fn insert_id_to_str(&mut self, key: EnvironmentInternalSymbolKey, value: String) {
        self.id_to_symbol.insert(key, value.clone());
        self.symbol_to_id.insert(value, key);
    }
    pub fn get_new_id(&mut self) -> EnvironmentInternalSymbolKey {
        self.unique_keys += 1;
        self.unique_keys
    }
}
